/*
 * Copyright (C) 2000  Transmeta Corporation
 *
 * written by Claudio Fleiner <cfl@transmeta.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <sys/mount.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <getopt.h> 
#include <zlib.h> 
#include <memory.h> 
#include <signal.h>

// make linux/loop.h happy with glibc 2.2
// remove this when loop.h has been fixed
#ifdef __dev_t_defined
#define dev_t dev_t
#endif

#include <linux/loop.h>
//#include <linux/fs.h>
#ifndef NO_DIRECT_MOUNT
#include <sys/ioctl.h>
#include <linux/blkpg.h>
#endif


#define xxDEBUG 

/* maximum length of mount point path */
#define MOUNTPOINT_LENGTH (128-16)

/* magic number */
#define MAGIC 0xeab85d45
#define CRAMFS_MAGIC		0x28cd3d45

/* signature string */
#define SIGNATURE "MCompressed"

/* current packcramfs version */
#define VERSION 2

#define SB_SIZE(sb)  (sizeof(struct super_block)+(((sb)->images)-1)*sizeof(struct image_struct))

/* blocksize used to write second superblock */
#define SB_BLOCK      512
#define NEXT_BLOCK(x) (((x) & ~(SB_BLOCK-1))+SB_BLOCK)
#define NEXT_FBLOCK(x) (((x) & ~(sb->blocksize-1))+sb->blocksize)
#define START(sb)     (NEXT_FBLOCK(NEXT_BLOCK(SB_SIZE(sb))+SB_SIZE(sb)))
#define END(sb)       ((sb)->size-1)

/* some useful macros */
#define PARTITION_ACTIVE(img) ((img).active)
#define PARTITION_EMPTY(img) ((img).start==0)
#define PARTITION_USED(img) !PARTITION_EMPTY(img)

#define MC_READ   1
#define MC_UPDATE 3
#define MC_CREATE 4

typedef unsigned int u32;
typedef unsigned int u16;
typedef unsigned char u8;

#define FATAL_ERROR         128
#define NON_FATAL_ERROR       0
#define NEED_FORCE         (1|FATAL_ERROR)
#define OUT_OF_PARTITIONS  (2|FATAL_ERROR)
#define CORRUPT_FILES      (3|FATAL_ERROR)
#define CORRUPT_SUPERBLOCK (4|FATAL_ERROR)
#define WRITE_ERROR        (5|FATAL_ERROR)
#define PARTITION_TOO_SMALL (6|FATAL_ERROR)
#define ILLEGAL_ARGUMENT   (7|FATAL_ERROR)
#define BUSY               (8|FATAL_ERROR)
#define CHECK_ERROR        (9|FATAL_ERROR)
#define INTERNAL_ERROR    (10|FATAL_ERROR)

#define NEED_CHECK         (1|NON_FATAL_ERROR)

#define SUCCESS             0


/* TODO should include cramfs.h instead for those defines and three structures */

#define CRAMFS_FLAG_FSID_VERSION_2      0x00000001      /* fsid version #2 */
#define CRAMFS_FLAG_HOLES               0x00000100      /* support for holes */
#define CRAMFS_FLAG_WRONG_SIGNATURE     0x00000200      /* !CRAMFS_SIGNATURE */
#define CRAMFS_FLAG_SHIFTED_ROOT_OFFSET 0x00000400      /* shifted root fs */  

// where the loopdevices are created, including the beginning of the filename
#define LOOPPATH "/tmp/loop"
// where disks and partitions are created
#define DISKPATH "/tmp/dsk"


struct fs_info {
       u32 crc;
       u32 edition;
       u32 blocks;
       u32 files;
};

/*
 * Reasonably terse representation of the inode data.
 */
struct cramfs_inode {
        u32 mode:16, uid:16;
        /* SIZE for device files is i_rdev */
        u32 size:24, gid:8;
        /* NAMELEN is the length of the file name, divided by 4 and
           rounded up.  (cramfs doesn't support hard links.) */
        /* OFFSET: For symlinks and non-empty regular files, this
           contains the offset (divided by 4) of the file data in
           compressed form (starting with an array of block pointers;
           see README).  For non-empty directories it is the offset
           (divided by 4) of the inode of the first file in that
           directory.  For anything else, offset is zero. */
        u32 namelen:6, offset:26;
};           

/*
 * Superblock information at the beginning of the FS.
 */
struct cramfs_super {
       u32 magic;              /* 0x28cd3d45 - random number */
       u32 size;               /* length in bytes */
       u32 flags;              /* 0 */
       u32 future;             /* 0 */
       u8 signature[16];       /* "Compressed ROMFS" */
       struct fs_info fsid;    /* unique filesystem info */
       u8 name[16];            /* user-defined name */
       struct cramfs_inode root;       /* Root inode data */
};

struct image_struct {
        u32 start; /* start of partition */
        u32 len;   /* length of partition in bytes */
        u32 edition; /* edition number (defined by server) */
        u8 active;        /* true if active and mounted */
        u8 reserve_1;
        u8 reserve_2;
        u8 reserve_3;
        u8 mountpoint[MOUNTPOINT_LENGTH]; /* mount point */
};

struct super_block {
        u32 magic;      /* 0xeab85d45  random number */
        u32 size;       /* overall size of partition */
        u16 images;     /* number of images supported (default 32) */
        u16 blocksize;  /* size of blocks */
        u32 crc;        /* crc for image table */

        u32 count;      /* incremented whenever the superblock changes */
        u16 version;    /* packcramfs version, currently 1 */
        u8 loopoffset;  /* offset of loopdevices */
        u8 future8;     /* for future use */
        u32 future[3];  /* for future use */
        char signature[12]; /* "Multi Compressed" */
        struct image_struct image[1];
};

struct cramfs_file {
        FILE *file;
        u32 start;
        u32 size;
        struct cramfs_super super;
        u8 *mountpoint;
        u8 *filename;
        int partition; // temporary to store partition where it may be written
        int hole; // temporary to store position where it may be written
};

struct compress_image {
        int partition;
        int start;
        int size;
};

struct partition {
        unsigned char boot;
        unsigned char start_head;
        unsigned char start_sector;
        unsigned char start_track;
        unsigned char os;
        unsigned char end_head;
        unsigned char end_sector;
        unsigned char end_track;
        unsigned int start;
        unsigned int len;
};

#define PARTITION_OFFSET 0x1BE



static void die(int exitcode,char *msg) {
        fprintf(stderr,"packcramfs: %s\n",msg);
        exit(exitcode);
}

static int quiet=0;
static int progress_bar=0;
static int error_code=SUCCESS;


static void usage(char *e) {
        FILE *f=stdout;
        if(e!=NULL) {
                f=stderr;
                fprintf(f,"packcramfs: %s\n",e);
        }
#ifndef NO_USAGE_MSG
        fprintf(f,"packcramfs -c|--create [-i|--images n] [-b|--blocksize blocksize]\n"
                  "                       [-l|--loop n] (blocckdevice|file) blocks\n"
                  "           create new multi cramfs partition with n partitions (default 16),\n"
                  "           and the given number of blocks of the given size (default block\n"
                  "           size is 1024). The loopoffset defines the first loopdevice that\n"
                  "           is used to mount images\n"
                  "packcramfs -p|--print [-a|--all] [-t|--terse] (blockdevice|file)\n"
                  "           print information about the active images or all images [-a]\n"
                  "packcramfs -D|--deactivate (blockdevice|file) n\n"
                  "           Deactivate image n (it is not deleted). It will also unmount all\n"
                  "           partitions before deactivating the partition.\n"
                  "packcramfs -D|--deactivate (blockdevice|file) mountpoint\n"
                  "           Deactivate image using the given mountpoint. It will also unmount\n"
                  "           all partitions before deactive the partition.\n"
                  "packcramfs -s|--store [-f|--force] [--append] (blockdevice|file)\n"
                  "                                        mountpoint file [mountpoint file ...]\n"
                  "           store new cramfs file image without disturbing the others unless\n"
                  "           -f is given. -f will also unmount all partitions. --append will\n"
                  "           append the partitions to the end of the file/device.\n"
                  "packcramfs -x|--extract (blockdevice|file) (mountpoint|imagenumber) [filename]\n"
                  "           Extract a given mountpoint (if there is more than one the active\n"
                  "           one is used, if none is active the first one) to the fil or stdout.\n"
                  "packcramfs -A|--activate (blockdevice|file) mountpoint [mountpoint ...]\n"
                  "           activate mountpoints after unmounting all partitions.\n"
                  "packcramfs -r|--restore (blockdevice|file)\n"
                  "           restore a partition by reactivating all deactivated images which\n"
                  "           where either deleted with -d or replaced with -s. Only one or more\n"
                  "           -d or one -s can be restored and it -r cannot be used if --force\n"
                  "           was used. It will unmount all partitions before restoring.\n"
                  "packcramfs -m|--mount (blockdevice|file) root\n"
                  "           Mount all images starting with /\n"
                  "packcramfs -u|--umount (blockdevice|file)\n"
                  "           unmount everything\n"
                  "packcramfs -C|--check [-a|--all] [-w|--write] [--resize] [-I|--image] [--fast]\n"
                  "                                                           (blockdevice|file)\n"
                  "           Check active (or all) partitions for CRC errors. Works even if the\n"
                  "           partitions are mounted. In the case of an error one should\n"
                  "           unmount everything and use -w|--write to fix the error or restore\n"
                  "           the previous version using --restore. If --resize is used with a\n"
                  "           block device the size of the filesystem will be changed to use the\n"
                  "           complete device, if it is used with a file it will truncate the file\n"
                  "           to make it as small as possible (without moving partitions). Using \n"
                  "           -I means that a single cramfs image instead of a complete packed \n"
                  "           partitions must be checked. --fast does not check the images but\n"
                  "           only the partition table\n"
                  "packcramfs -z|--compress (blockdevice|file)\n"
                  "           compress the archive, remove all non active partitions. Will first\n"
                  "           unmount all partitions.\n"
                  "\n"
                  "OPTIONS:   -q|--quiet         print less information\n"
                  "           -a|--all           print all partitions/images\n"
                  "           -f|--force         force changes\n"
                  "           -P|--progress-bar  print progress bar\n"
                  "\n"
                  "EXIT CODES:  0 on success\n"
                  "             errors below 128 are non fatal error messages, above 128 are fatal\n"
                  "\n"
                  "             1 if one superblock is corrupt and should be fixed. However, the\n"
                  "               program could still finish successfully.\n"
                  "\n"
                  "           129 if the option --force should be used\n"
                  "           130 if there are not enough partitions left\n"
                  "           131 if the cramfs files passed as argument are corrupt\n"
                  "           132 if the superblock is corrupt\n"
                  "           133 if there was a write error\n"
                  "           134 if there is not enough space on the device\n"
                  "           135 if the arguments are wrong\n"
                  "           136 if there was a problem mounting/unmounting due to a busy file\n"
                  "               or device\n"
                  "           137 if check removed some partitions (or recommends to remove them)\n"
                  "               because of errors\n"
                  "           138 if there was an internal error (out of memory, ...)\n"
                );
#endif /* NO_USAGE_MSG */
        exit(e==NULL?0:ILLEGAL_ARGUMENT);
}

struct holes_struct {
        u32 start;
        u32 end;
        u32 unused;
};

// list of signals to turn on and off 
static int sigs[] = { SIGHUP,SIGINT,SIGQUIT,SIGTERM,SIGALRM,SIGUSR1,SIGUSR2,SIGCHLD };
typedef void(*sigtype)(int);
static sigtype sigsorig[sizeof(sigs)/sizeof(int)] = { NULL };
static int siglevel=0;

static int compare_holes(struct holes_struct *a,struct holes_struct *b) {
        if(a->unused<b->unused) return 1;
        if(a->unused>b->unused) return -1;
        return 0;
}

static void ignore_signals(void) {
        int i;
        int init=0;
        if(siglevel++) return;
        init=sigsorig[0]==NULL;
        for(i=0;i<sizeof(sigs)/sizeof(int);i++)
                if(init) sigsorig[i]=signal(sigs[i],SIG_IGN);
                else signal(sigs[i],SIG_IGN);
}

static void enable_signals(void) {
        int i;
        if(--siglevel) return;
        for(i=0;i<sizeof(sigs)/sizeof(int);i++)
                signal(sigs[i],sigsorig[i]);
}

/*
 * Find all holes (unused space).
 */
static struct holes_struct *find_holes(struct super_block *sb) {
        int i,j,next;
        // always allocate at least one more than necessary
        // in order to make room for an append hole
        struct holes_struct *h;
        h=(struct holes_struct *)malloc((sb->images+3)*sizeof(struct holes_struct));
        memset(h,(sb->images+3)*sizeof(struct holes_struct),0);
        h[0].start=START(sb);
        h[0].end=END(sb);
        //printf("FIRST HOLE: 0x%08x - 0x%08x\n",h[0].start,h[0].end);
        next=1;
        for(i=0;i<sb->images;i++) {
                if(PARTITION_USED(sb->image[i])) {
                        int s=sb->image[i].start;
                        for(j=0;j<next;j++) {
                                if(h[j].start<=s && h[j].end>s) {
                                        if(s==h[j].start) {
                                                h[j].start=s+sb->image[i].len;
                                        } else if(h[j].end==s+sb->image[i].len) {
                                                h[j].end=s-1;
                                        } else {
                                                h[next].end=h[j].end;
                                                h[next].start=s+sb->image[i].len;
                                                h[j].end=s-1;
                                                next++;
                                        }
                                        break;
                                }
                        }
                }
        }
        for(i=0;h[i].start;i++) {
                h[i].unused=h[i].end-h[i].start+1;
                if(!h[i].unused) h[i].start=h[i].end=0;
        }
        qsort(h,next,sizeof(*h),(int(*)(const void*,const void*))compare_holes);
        return h;
}

static int progress_current=0;
static int progress_top=0;
static char progress_char='*';
#define PRLEN 65
static void init_progress_bar(char *title,char c,int top) {
        char b[PRLEN|1];
        if(!progress_bar) return;
        memset(b,'-',65);
        b[PRLEN]=0;
        printf("%-10s: |%s|\r%-10s: |",title,b,title);
        fflush(stdout);
        progress_top=top;
        progress_char=c;
        progress_current=0;
}

static void set_progress_bar(int c) {
        int o,n;
        if(c<=progress_current || !progress_bar) return;
        o=progress_current*PRLEN/progress_top;
        n=c*PRLEN/progress_top;
        if(n<=o) return;
        while(o++<n) putchar(progress_char);
        fflush(stdout);
        progress_current=c;
}

static void reset_progress_bar() {
        if(!progress_bar) return;
        printf("\r%79s\r","");
        fflush(stdout);
}

static void print_superblock(struct super_block *sb,int all,int terse) {
        int i,active=0,used=0;
        u32 free,unused;
        struct holes_struct *h;

        unused=free=sb->size-START(sb);
        for(i=0;i<sb->images;i++) {
                if(sb->image[i].active) {
                        active++;
                        free-=sb->image[i].len;
                }
                if(sb->image[i].len) {
                        unused-=sb->image[i].len;
                        used++;
                }
        }

        h=find_holes(sb);

        if(!quiet) {
                printf("magic:                      0x%08x\n",sb->magic);
                printf("version:                    %d\n",sb->version);
                printf("count:                      %d\n",sb->count);
                printf("signature:                  %s\n",sb->signature);
                printf("overall size:               %d\n",sb->size);
                printf("block size:                 %d\n",sb->blocksize);
                printf("loop offset:                %d\n",sb->loopoffset);
                printf("total number of partitions: %d\n",sb->images);
                printf("active partitions:          %d\n",active);
                printf("used partitions:            %d\n",used);
                printf("non active space:           %d\n",free);
                printf("unused space:               %d\n",unused);
                printf("biggest hole:               %d\n",h[0].unused);
        }

        for(i=0;i<sb->images;i++) {
                if(sb->image[i].active || all) {
                        if(terse) {
                                printf("%d %s %u\n",i, sb->image[i].mountpoint,
                                        sb->image[i].edition);
                        } else {
                                printf("[%02d]: %c %-19s (edition %u, start=0x%08x, size=%d bytes)\n",i,
                                        sb->image[i].active?'*':' ',sb->image[i].mountpoint,
                                        sb->image[i].edition,sb->image[i].start,sb->image[i].len);
                        }
                }
        }
}


/*
 * Calculate superblock crc. It will not change it though.
 */
static u32 superblock_crc(struct super_block *sb) {
        u32 original=sb->crc;
        u32 crc;
        sb->crc=0;
        crc=crc32(0L, Z_NULL, 0);
        crc=crc32(crc,(char*)sb,SB_SIZE(sb));
        sb->crc=original;
        return crc;
}

/*
 * Check superblock crc and return 1 if ok, 0 if error
 */
static int check_superblock_crc(struct super_block *sb) {
        return sb->crc==superblock_crc(sb);
}

/*
 * Read superblock, return NULL on error.
 */
static struct super_block *read_sb(FILE *f) {
        struct super_block *sb;
        int size,got;

//#define R(x) { fprintf(stderr,x "\n");return NULL; }
#define R(x) { return NULL; } 

        size=sizeof(struct super_block);
        sb=(struct super_block *)malloc(size);
        if(!sb) die(INTERNAL_ERROR,"not enough memory");
        got=fread(sb,1,size,f);
        if(size!=got) R("reading problem\n");
        if(sb->magic!=MAGIC) R("magic wrong");
        if(sb->version!=VERSION) R("version wrong");
        size=SB_SIZE(sb);
        if(size>sb->size || size<sizeof(struct super_block)) R("size wrong");
        sb=(struct super_block *)realloc(sb,size);
        if(!sb) R("memory problem");

        got=fread(((char *)sb)+sizeof(struct super_block),1,size-sizeof(struct super_block),f);
        if(size!=got+sizeof(struct super_block)) R("reading size problem");
        if(sb->magic!=MAGIC) R("magic problem");
        if(strcmp(sb->signature,SIGNATURE)) R("signature problem");
        if(!check_superblock_crc(sb)) R("corrupt block");

        return sb;
}

/*
 * read second superblock assuming that the first one is corrupt.
 */
static struct super_block *find_second_superblock(FILE *f) {
        int pos;
        u32 magic;
        struct super_block *sb;
        // first lets find the second superblock.
        // It starts at an SB_BLOCK boundary, within the first 32 blocks.
        // We look for the correct magic number.
        for(pos=SB_BLOCK;pos<=SB_BLOCK*32;pos++) {
                fseek(f,pos,SEEK_SET);
                if(fread(&magic,1,sizeof(magic),f)!=sizeof(magic)) 
                        die(CORRUPT_SUPERBLOCK,"filesystem corrupt (cannot find either superblock)");
                if(magic==MAGIC) {
                        if(!quiet) printf("found second superblock at %d\n",pos);
                        fseek(f,pos,SEEK_SET);
                        sb=read_sb(f);
                        if(sb==NULL) die(CORRUPT_SUPERBLOCK,"both superblocks are corrupt");
                        return sb;
                }
        }
        die(CORRUPT_SUPERBLOCK,"first superblock corrupt, second superblock vanished");
        return NULL;
}

static struct super_block *check_second_superblock(FILE *f,struct super_block *sb) {
        u32 size,pos;
        struct super_block *sb2;
        // the second superblock must start after the first one
        // aligned to SB_BLOCK.
        size=SB_SIZE(sb);
        pos=(size & ~(SB_BLOCK-1))+SB_BLOCK;
        fseek(f,pos,SEEK_SET);
        sb2=read_sb(f);
        if(sb2==NULL) {
                int w;
                // corrupt or different second superblock.
                // We will fix it now. We assume that the first
                // one is the correct one (the first one is ALWAYS
                // written first!
                fseek(f,pos,SEEK_SET);
                size=SB_SIZE(sb);
                w=fwrite(sb,1,size,f);
                if(w<=0) {
                        // probably writeprotected
                        fprintf(stderr,"the second superblock is corrupt, use --check --write to fix it\n");
                        error_code=NEED_CHECK;
                        return sb;
                } else if(w!=size) {
                        fprintf(stderr,"write size error: wrote %d instead of %d\n",w,size);
                        die(WRITE_ERROR,"error fixing second superblock");
                }
                fseek(f,pos,SEEK_SET);
                sb2=read_sb(f);
                if(sb2==NULL) die(WRITE_ERROR,"error writing second superblock");
        } else if(memcmp(sb,sb2,size)) {
                // lets see which one is the older one, that one gets
                // overwritten.
                struct super_block *sb3,*sb4;
                int place,w;
                if(sb->count>sb2->count) {
                        place=pos;
                        sb3=sb;
                } else {
                        place=0;
                        sb3=sb2;
                }
                size=SB_SIZE(sb3);
                fseek(f,place,SEEK_SET);
                w=fwrite(sb3,1,size,f);
                if(w<=0) {
                        // probably writeprotected
                        fprintf(stderr,"one superblock is out of date, use --check --write to fix it\n");
                        error_code=NEED_CHECK;
                        return sb3;
                } else if(w!=size) die(CORRUPT_SUPERBLOCK,"error fixing superblock");
                fseek(f,place,SEEK_SET);
                sb4=read_sb(f);
                if(sb4==NULL || memcmp(sb3,sb4,size)) die(CORRUPT_SUPERBLOCK,"cannot fix corrupted superblock");
                return sb3;
        }
        return sb;
}

/*
 * read superblock
 */
static struct super_block *read_superblock(FILE *f) {
        struct super_block *sb,*sbc;
        int size,w;

        fseek(f,0,SEEK_SET);
        sb=read_sb(f);
        if(sb==NULL) {
                sb=find_second_superblock(f);
                // need to write first superblock
                size=SB_SIZE(sb);
                fseek(f,0,SEEK_SET);
                w=fwrite(sb,1,size,f);
                if(w<=0) {
                        // probably readonly, we print an error message
                        // but continue anyway
                        fprintf(stderr,"first superblock corrupt, second one seems ok\nuse option --check --write to fix\n");
                        error_code=NEED_CHECK;
                        return sb;
                } else if(w!=size) {
                        fprintf(stderr,"write size error: %d instead of %d\n",w,size);
                        die(WRITE_ERROR,"write error while fixing corrupt superblock");
                }
                fseek(f,0,SEEK_SET);
                sbc=read_sb(f);
                if(sbc==NULL) die(WRITE_ERROR,"cannot fix corrupted superblock");
                if(memcmp(sbc,sb,size)) die(WRITE_ERROR,"error while fixing superblock");
                return sb;
        } else return check_second_superblock(f,sb);
}

/*
 * write superblock. Dies if there is an error.
 * File must be opened as 'UPDATE'
 * returns size of superblock.
 */
static int write_superblock(FILE *f,struct super_block *sb) { 
        int size,pos;
        struct super_block *sbr;
        // printf("writing superblock\n");
        // print_superblock(sb,1);
        sb->count++; // increase internal version
        sb->version=VERSION;
        sb->crc=superblock_crc(sb);
        size=SB_SIZE(sb);
        ignore_signals();
        fseek(f,0,SEEK_SET);
        if(size!=fwrite(sb,1,size,f)) die(PARTITION_TOO_SMALL,"partition too small");
        fflush(f);
        
        // write second superblock
        pos=(size & ~(SB_BLOCK-1))+SB_BLOCK;
        fseek(f,pos,SEEK_SET);
        if(size!=fwrite(sb,1,size,f)) die(PARTITION_TOO_SMALL,"partition too small");

        sbr=read_superblock(f);
        if(memcmp(sbr,sb,size)) die(CORRUPT_SUPERBLOCK,"error while updating superblock");
        enable_signals();
        
        return size;
}

/*
 * open device, mode can be READ, WRITE, UPDATE
 */
static FILE *opendev(u8 *dev,int mode,struct super_block **sb) {
        FILE *f=NULL;
        int fd,flags=0;
        switch(mode) {
        case MC_READ: f=fopen(dev,"r");break;
        case MC_CREATE: 
              flags|=O_CREAT|O_TRUNC;
              // break missing on purpose
        case MC_UPDATE: 
              flags|=O_RDWR|O_SYNC;
              fd=open(dev,flags,0600);
              if(fd<0) {
                      perror("open");
                      die(WRITE_ERROR,"cannot open device");
              }
              f=fdopen(fd,"r+");
              break;

        default: die(INTERNAL_ERROR,"internal error 12");
        }
        if(f==NULL) die(INTERNAL_ERROR,"cannot open device");
        if(sb!=NULL && mode!=MC_CREATE)
                *sb=read_superblock(f);
        return f;
}

/*
 * print the info contained in the super block
 */
static void print(u8 *dev,int all,int terse) {
        struct super_block *sb;

        opendev(dev,MC_READ,&sb);
        if(!quiet) printf("Multi Compressed partition, reading data from %s\n",dev);
        print_superblock(sb,all,terse);

}

/*
 * delete a partition entry
 */
static void delete_partition(struct image_struct *img) {
        img->active=0;
        img->start=0;
        img->len=0;
        img->edition=0;
        memset(img->mountpoint,0,MOUNTPOINT_LENGTH);
}

/*
 * Create a new image. Deletes everything on that partition.
 */
static void create(u8 *dev,u32 blocks,u32 blocksize,u32 images,u32 loopoffset) {
        int i;
        struct super_block *sb,*sbr;
        int size,pos;
        FILE *f;
        size=sizeof(struct super_block)+(images-1)*sizeof(struct image_struct);
        sb=(struct super_block *)malloc(size);
        memset(sb,0,size);
        if(!sb) die(INTERNAL_ERROR,"not enough memory");
        sb->magic=MAGIC;
        sb->size=blocks*blocksize;
        sb->images=images;
        sb->blocksize=blocksize;
        sb->crc=0;
        sb->loopoffset=loopoffset;
        strcpy(sb->signature,SIGNATURE);
        for(i=0;i<images;i++)
                delete_partition(&sb->image[i]);
        f=opendev(dev,MC_CREATE,NULL);
        if(write_superblock(f,sb)!=size) die(INTERNAL_ERROR,"internal error 221");

        // write superblock a second time at the beginning of the next block
        // of size 512.

        pos=(size & ~(SB_BLOCK-1))+SB_BLOCK;
        fseek(f,pos,SEEK_SET);
        if(write_superblock(f,sb)!=size) die(INTERNAL_ERROR,"internal error 222");

        // now seek to the end and write one byte 
        // (creates a file with holes!)
        fseek(f,sb->size-1,SEEK_SET);
        i=0;
        if(fwrite(&i,1,1,f)!=1) die(WRITE_ERROR,"cannot write to the end of the device");

        fclose(f);

        f=opendev(dev,MC_READ,&sbr);
        if(memcmp(sb,sbr,size)) die(WRITE_ERROR,"error while writing");

        if(!quiet) {
                printf("created superblock of size %d bytes, with %d partitions\n",size,images);
                print(dev,0,0);
        }
}

#ifndef NO_DIRECT_MOUNT
/* code from util-linux addpart.c */
static int add_part(int fd, int pno, long start, long length) {
	struct blkpg_ioctl_arg a;
	struct blkpg_partition p;

	p.pno = pno;
	p.start = 512LL * (long long)start;
	p.length = 512 * length;
	p.devname[0] = 0;
	p.volname[0] = 0;
	a.op = BLKPG_ADD_PARTITION;
	a.flags = 0;
	a.datalen = sizeof(p);
	a.data = &p;

	if (ioctl(fd, BLKPG, &a) == -1) {
                printf("error: %d\n",errno);
		perror("BLKPG");
		return -1;
	}
	    
	return 0;
}

/* code from util-linux delpart.c */
static int del_part(int fd, int pno) {
	struct blkpg_ioctl_arg a;
	struct blkpg_partition p;

	p.pno = pno;
	p.start = 0;
	p.length = 0;
	p.devname[0] = 0;
	p.volname[0] = 0;
	a.op = BLKPG_DEL_PARTITION;
	a.flags = 0;
	a.datalen = sizeof(p);
	a.data = &p;

	if (ioctl(fd, BLKPG, &a) == -1)  return -1;
	    
	return 0;
}
#endif

/*
 * find the root of all mounts
 */
static char *find_root(FILE *disk,struct super_block *sb) {
        FILE *f;
        char line[2*PATH_MAX];
        int i,root;
        char loopdev[strlen(LOOPPATH)+5];
        char diskpath[strlen(DISKPATH)+15];
        char *p;
        struct stat statbuf;
        int major=-1;

        if(!fstat(fileno(disk), &statbuf) && S_ISBLK(statbuf.st_mode)) 
                major=statbuf.st_rdev>>8;

        f=fopen("/proc/mounts","r");
        if(f==NULL) die(INTERNAL_ERROR, "unable to open /proc/mounts");
        // find root partition
        for(i=root=0;root<sb->images;root++)
                if(sb->image[root].active && !strcmp(sb->image[root].mountpoint,"/")) {
                        i=1;
                        break;
                }

        if(!i) return NULL; // no root partition, ergo nothing mounted
        sprintf(loopdev,LOOPPATH "%d ",root+sb->loopoffset);
        sprintf(diskpath,DISKPATH "_%x_%x",major,root+sb->loopoffset+5);

        while(!feof(f)) {
                fgets(line,2*PATH_MAX,f);
                if(!strncmp(line,loopdev,strlen(loopdev)) || !strncmp(line,diskpath,strlen(diskpath))) {
                        p=line;
                        while(*p!=' ' && *p) p++;
                        p++;
                        for(i=0;p[i];i++)
                                if(p[i]==' ') {
                                        p[i]=0;
                                        break;
                                }
                        return (char*)strdup(p);
                }
        }
        return NULL;
}

#ifndef NO_DIRECT_MOUNT
#define BUFCHECKLEN 10000
#undef R
#ifdef NO_LOOPBACK
# define R(x) die(ILLEGAL_ARGUMENT,x);
#else
# define R(x) return 0;
#endif
/*
 * Return NULL if direct mounting is not possible,
 * aborts if the NO_LOOPBACK option is on.
 */
static int check_for_direct_mount(FILE *f,struct super_block *sb,FILE **diskfd,unsigned long *offset) {
        int major,minor,i;
        unsigned long off;
        char disk[strlen(DISKPATH)+15];
        struct stat statbuf;
        unsigned char buf1[BUFCHECKLEN],buf2[BUFCHECKLEN];
        struct partition p;
        FILE *d;

        if(fstat(fileno(f), &statbuf)) 
                R("cannot stat() device");
        if(!S_ISBLK(statbuf.st_mode)) 
                R("need a block device!");
        major=statbuf.st_rdev>>8;
        minor=statbuf.st_rdev&0xff;
        if(minor<1 || minor>4) 
                R("need primary partition");

        sprintf(disk,DISKPATH "_%x_%x",major,0);
        if(stat(disk, &statbuf)) {
                mknod(disk,S_IFBLK|0600,(major<<8));
                if(!stat(disk, &statbuf) == 0) 
                        R( "cannot create disk device");
        }
        d=fopen(disk,"r");
        if(d==NULL) R("cannot open disk for reading");
        fseek(d,PARTITION_OFFSET+sizeof(struct partition)*(minor-1),SEEK_SET);
        i=fread(&p,sizeof(p),1,d);
        if(i!=1) R("cannot read from disk");

        // sanity check: read 10KB from partition and from disk and compare
        fseek(f,0,SEEK_SET);
        i=fread(buf1,1,BUFCHECKLEN,f);
        if(i!=BUFCHECKLEN) R("cannot read from partition");

        // make sure there is no overflow when seeking
        i=512;
        off=p.start;
        while(off<0x40000000 && i>1) { off<<=1;i>>=1; }
        fseek(d,0,SEEK_SET);
        while(i--) fseek(d,off,SEEK_CUR);
        i=fread(buf2,1,BUFCHECKLEN,d);
        if(i!=BUFCHECKLEN) R("cannot read from disk");

        if(memcmp(buf1,buf2,BUFCHECKLEN))
                R("disk and partition do not match??");

        *diskfd=d;
        *offset=p.start;

        return 1;
}
#endif

struct umount_data {
        int part;
        char *name;
};

/*
 * sort mount data by length of mountpath
 */
static int compare_rlen(struct umount_data *a,struct umount_data *b) {
        int la=strlen(a->name);
        int lb=strlen(b->name);
        if(la>lb) return -1;
        if(la==lb) return 0;
        return 1;
}

/*
 * make sure none of the partitions are still mounted.
 */
static void do_umount_all(FILE *f,struct super_block *sb) {
        struct umount_data *ud;
        struct loop_info loopinfo;
        char loopdev[strlen(LOOPPATH)+5];
        char diskpath[strlen(DISKPATH)+15];
        int i,fd;
        char *root;
        char dir[PATH_MAX];
        int use_direct;
        FILE *disk;
        unsigned long offset;
        int major;
        struct stat statbuf;

        root=find_root(f,sb);
#ifndef NO_DIRECT_MOUNT
        use_direct=check_for_direct_mount(f,sb,&disk,&offset);
#endif
        
        if(!fstat(fileno(f), &statbuf) && S_ISBLK(statbuf.st_mode)) 
                major=statbuf.st_rdev>>8;

        // need to unmount in correct order to avoid spurious busy errors
        ud=(struct umount_data *)malloc(sizeof(struct umount_data)*sb->images);
        for(i=0;i<sb->images;i++) {
                ud[i].part=i;
                ud[i].name=sb->image[i].mountpoint;
        }
        //printf("umount, use_direct=%d, root=%s\n",use_direct,root);
        qsort(ud,sb->images,sizeof(*ud),(int(*)(const void*,const void*))compare_rlen);
        for(i=0;i<sb->images;i++) {
#ifndef NO_DIRECT_MOUNT
                if(use_direct) { 
                        sprintf(diskpath,DISKPATH "_%x_%x",major,5+ud[i].part+sb->loopoffset);
                        if(root!=NULL) {
                                strcpy(dir,root);
                                if(strcmp(sb->image[ud[i].part].mountpoint,"/")) {
                                        if(sb->image[ud[i].part].mountpoint[0]!='/') strcat(dir,"/");
                                        strcat(dir,sb->image[ud[i].part].mountpoint);
                                }
                                if(umount(dir)) {
                                        if(errno==EBUSY) {
                                                fprintf(stderr,"unmounting %s failed\n",dir);
                                                perror(dir);
                                                die(BUSY,"cannot umount some of the filesystems, aborting!");
                                        }
                                } else {
                                        if(!quiet) printf("unmounting %s\n",dir);
                                }
                        }
                        del_part(fileno(disk),5+ud[i].part+sb->loopoffset);
                } else  
#endif
                        {
#ifdef NO_LOOPBACK
                        die(ILLEGAL_ARGUMENT,"need loopback support");
#else
                        sprintf(loopdev,LOOPPATH "%d",ud[i].part+sb->loopoffset);
                        if(root!=NULL) {
                                strcpy(dir,root);
                                if(strcmp(sb->image[ud[i].part].mountpoint,"/")) {
                                        if(sb->image[ud[i].part].mountpoint[0]!='/') strcat(dir,"/");
                                        strcat(dir,sb->image[ud[i].part].mountpoint);
                                }
                                if(umount(dir)) {
                                        if(errno==EBUSY) {
                                                fprintf(stderr,"unmounting %s failed\n",dir);
                                                perror(dir);
                                                die(BUSY,"cannot umount some of the filesystems, aborting!");
                                        }
                                } else if(!quiet) printf("unmounting %s\n",dir);
                        }
                        if((fd=open(loopdev, O_RDONLY))>=0) {
                                if (ioctl(fd, LOOP_GET_STATUS, &loopinfo) == -1) {
                                        if (errno == ENXIO) {   /* probably free */
                                                close(fd);
                                        } else 
                                                die(BUSY,"loopback device cannot be used");
                                }
                                ioctl(fd, LOOP_CLR_FD, 0);
                                close(fd);
                        }
#endif
                }
        }
}


/*
 * sort cramfs_file array by size, biggest file first.
 */
static int compare_size(struct cramfs_file *a,struct cramfs_file *b) {
        if(a->size>b->size) return -1;
        if(a->size==b->size) return 0;
        return 1;
}

/*
 * check crc for cramfs image that is part of a larger file.
 */
static int check_cramfs_crc(FILE *f,u32 start,u32 length) {
        u32 crc,orig;
        int i,sum,size;
        struct cramfs_super super;
        char buf[1024];

        fseek(f,start,SEEK_SET);
        size=fread(&super,1,sizeof(super),f);

        crc=crc32(0L, Z_NULL, 0);
        orig=super.fsid.crc;
        super.fsid.crc=crc;
        crc=crc32(crc,(char*)&super,sizeof(super));
        sum=size;
        init_progress_bar("checking",'+',length);
        while((i=fread(buf,1,1024,f))>0) {
                if(sum+i>length) i=length-sum;
                crc=crc32(crc,buf,i);
                sum+=i;
                set_progress_bar(sum);
                if(sum==length) break;
        }
        reset_progress_bar();
        
        return crc==orig && sum==length;
}

/*
 * read a cramfs file
 */
static void read_cramfs(struct cramfs_file *cf,char *mp,char *f,int check) {
        int size,offset=0;
        struct stat statbuf;
        if(!quiet) printf("read_cramfs(%s, %s)\n",mp,f);
        cf->file=fopen(f,"r");
        if(!cf->file) {
                fprintf(stderr,"cannot open file %s\n",f);
                die(CORRUPT_FILES,"aborting");
        }
        cf->mountpoint=mp;
        cf->filename=f;
        size=fread(&cf->super,1,sizeof(cf->super),cf->file);
        if(size!=sizeof(cf->super)) die(CORRUPT_FILES,"error reading cramfs image");
        if(cf->super.magic!=CRAMFS_MAGIC) {
                fseek(cf->file,512,SEEK_SET);
                size=fread(&cf->super,1,sizeof(cf->super),cf->file);
                if(size!=sizeof(cf->super)) die(CORRUPT_FILES,"error reading cramfs image");
                offset=512;
        }
        fseek(cf->file,0,SEEK_END);
        if(fstat(fileno(cf->file), &statbuf)) die(INTERNAL_ERROR,"cannot stat() device");
        if(S_ISBLK(statbuf.st_mode)) cf->size=cf->super.size;
        else cf->size=ftell(cf->file);

        if(!cf->size) cf->size=cf->super.size; // probably a block device
        if(cf->super.size!=cf->size) {
                printf("cf->super.size=%d\ncf->size=%d\n",cf->super.size,cf->size);
                die(CORRUPT_FILES,"cramfs size field corrupt");
        }
        if(check) { 
                if(!(cf->super.flags&CRAMFS_FLAG_FSID_VERSION_2))
                        die(CORRUPT_FILES,"old cramfs image, need version 2");
                if(!check_cramfs_crc(cf->file,offset,cf->size-offset))
                        die(CORRUPT_FILES,"cramfs has crc error");
        }
        fseek(cf->file,offset,SEEK_SET);
}

/*
 * print info about a cramfs file
 */
static void print_image(u8 *dev) {
        struct cramfs_file cf;
        cf.size=0;
        
        read_cramfs(&cf,"",dev,0);
        printf("edition %u  files %u  blocks %u\n",cf.super.fsid.edition,cf.super.fsid.files,cf.super.fsid.blocks);
}


/*
 * overall size of holes (=free space)
 */
static u32 free_space(struct holes_struct *h) {
        int i;
        u32 total=0;

        for(i=0;h[i].start;i++)
                total+=h[i].unused;
        return total;
}

/*
 * overall size of new partitions.
 */
static u32 space_needed(struct cramfs_file *cf) {
        int i;
        u32 total=0;
        for(i=0;cf[i].file;i++) 
                total+=cf[i].size;
        return total;
}




/*
 * write a cramfs image.
 */
static void write_cramfs(FILE *f,struct super_block *sb,struct cramfs_file *cf,struct holes_struct *h) {
        char buf[1024];
        int i,j;
        assert(h->unused>=cf->size);
        cf->start=h->start;
        fseek(f,h->start,SEEK_SET);
        fseek(cf->file,0,SEEK_SET);
        j=0;
        init_progress_bar("writing",'*',cf->size);
        while((i=fread(buf,1,1024,cf->file))>0) {
                j+=i;
                if(fwrite(buf,1,i,f)!=i) 
                        die(INTERNAL_ERROR,"error while writing cramfs image");
                set_progress_bar(j);
        }
        reset_progress_bar(j);
        if(j!=cf->size) 
                die(INTERNAL_ERROR,"error: part of the cramfs image was not written");
        h->start+=cf->size;
        h->unused-=cf->size;
        sb->image[cf->partition].start=cf->start;
        sb->image[cf->partition].len=cf->size;
        sb->image[cf->partition].edition=cf->super.fsid.edition;
        sb->image[cf->partition].active=0;
        strcpy(sb->image[cf->partition].mountpoint,cf->mountpoint);
        /*
        for(i=0;i<sb->images;i++)
                if(!strcmp(sb->image[i].mountpoint,cf->mountpoint) && i!=cf->partition)
                        sb->image[i].active=0;
                        */
}

static int compare_start(struct compress_image *a,struct compress_image *b) {
        if(a->start>b->start) return 1;
        if(a->start==b->start) return 0;
        return -1;
}

/*
 * move a part of a file to an lower address.
 * Overlapping is ok.
 */
static void move_to_front(FILE *f,u32 old,u32 new,u32 size) {
        u32 s;
        int r;
        u32 copied;
        char buf[65536];

        if(new==old) return;
        s=size;
        if(s>65536) s=65536;

        copied=0;
        init_progress_bar("moving",'<',size);
        while(copied!=size) {
                if(copied+s>size) s=size-copied;
                fseek(f,old+copied,SEEK_SET);
                r=fread(buf,1,s,f);
                if(r<0) die(INTERNAL_ERROR,"read error");
                fseek(f,new+copied,SEEK_SET);
                if(fwrite(buf,1,r,f)!=r) die(WRITE_ERROR,"write error");
                copied+=r;
                set_progress_bar(copied);
        }
        reset_progress_bar();
}

/*
 * compress an image by removing all non active partitions
 * and moving everything to the beginning of the flash.
 */
static void compress_all(FILE *f,struct super_block *sb) {
        struct compress_image *ci;
        int i,n;
        u32 pos;

        do_umount_all(f,sb);
        ci=(struct compress_image *)malloc(sizeof(struct compress_image)*sb->images);
        for(n=i=0;i<sb->images;i++) {
                if(PARTITION_ACTIVE(sb->image[i])) {
                        ci[n].partition=i;
                        ci[n].start=sb->image[i].start;
                        ci[n++].size=sb->image[i].len;
                } else delete_partition(&sb->image[i]);
        }
        qsort(ci,n,sizeof(*ci),(int(*)(const void*,const void*))compare_start);

        pos=START(sb);
        for(i=0;i<n;i++) {
                int sigign=0;
                if(pos+ci[i].size>=ci[i].start) {
                        sb->image[ci[i].partition].active=0;
                        write_superblock(f,sb);
                        sigign=1;
                        ignore_signals();
                }
                move_to_front(f,ci[i].start,pos,ci[i].size);
                sb->image[ci[i].partition].start=pos;
                sb->image[ci[i].partition].active=1;
                pos+=ci[i].size;
                if(!check_cramfs_crc(f,sb->image[ci[i].partition].start,sb->image[ci[i].partition].len)) 
                        die(INTERNAL_ERROR,"write error: corrupt flash");
                write_superblock(f,sb);
                if(sigign) enable_signals();
        }
}

/*
 * compress an image by removing all non active partitions
 * and moving everything to the beginning of the flash.
 */
static void compress_mc(char *dev) {
        FILE *f;
        struct super_block *sb;
        f=opendev(dev,MC_UPDATE,&sb);
        compress_all(f,sb);
        print_superblock(sb,1,0);
}

/*
 * find free partitions for the given cramfs images and return 1 if it worked.
 * If force is true it will delete the old partitions with the same mount
 * point first (even if there would be enough free partitions)
 */
static int find_partitions(struct super_block *sb, struct cramfs_file *cf,int force) {
        int i,j;
        if(force) 
                for(i=0;i<sb->images;i++) 
                        for(j=0;cf[j].file;j++) 
                                if(!strcmp(sb->image[i].mountpoint,cf[j].mountpoint)) {
                                        delete_partition(&sb->image[i]);
                                        break;
                                }


        // delete all unused partitions
        for(i=0;i<sb->images;i++) 
                if(!PARTITION_ACTIVE(sb->image[i])) 
                        delete_partition(&sb->image[i]);

        // select unused partitions for new images
        for(j=i=0;i<sb->images && cf[j].file!=NULL;i++)
                if(PARTITION_EMPTY(sb->image[i])) 
                        cf[j++].partition=i;

        return !cf[j].file;
}

/*
 * Try to assign cramfs images to holes while optimizing the usage
 * of holes (current metric used: maximize geometric sum of remaining hole
 * sizes in blocks). Returns 1 if a solution was found.
 */
static int drop_cramfs_into_holes_r(u32 *best,struct cramfs_file *cf,struct holes_struct *h) {
        if(cf->file==NULL) {
                // end of recursion, check if this is better solution;
                u32 sum=0;
                int i;
                for(i=0;h[i].start;i++)
                        sum+=(h[i].unused>>10)*(h[i].unused>>10);
                if(sum>*best || *best==0xffffffff) {
                        *best=sum;
                        return 1;
                } 
                return 0;
        } else {
                int i,found=0;
                for(i=0;h[i].start;i++) {
                        if(h[i].unused>=cf->size) {
                                int last=cf->hole;
                                h[i].unused-=cf->size;
                                cf->hole=i;
                                if(!drop_cramfs_into_holes_r(best,cf+1,h))
                                        cf->hole=last;
                                else 
                                        found=1;
                                h[i].unused+=cf->size;
                        }
                }
                return found;
        }
}

/*
 * Try to assign cramfs images to holes while optimizing the usage
 * of holes (current metric used: maximize geometric sum of remaining hole
 * sizes in blocks). Returns 1 if a solution was found.
 */
static int drop_cramfs_into_holes(struct cramfs_file *cf,struct holes_struct *h) {
        u32 best=0xffffffff;
        return drop_cramfs_into_holes_r(&best,cf,h);
}

/*
 * Store several cramfs images (active ones are only moved around or deleted
 * [if replaced] if force==1)
 */
static int store(u8 *dev,char *argv[],int force,int append,int terse) {
        int i;
        int n=-1;
        FILE *f;
        int superblock_written=0;
        struct cramfs_file *cf;
        struct holes_struct *h;

        struct super_block *sb;
        while(argv[++n]);
        if(n&1) usage("the last mountpoint has no file?");
        n>>=1;

        f=opendev(dev,MC_UPDATE,&sb);
        if(force) do_umount_all(f,sb);

        cf=(struct cramfs_file*)malloc(sizeof(struct cramfs_file)*(n+1));
        if(!cf) die(INTERNAL_ERROR,"not enough memory");

        for(i=0;i<n;i++) {
                read_cramfs(cf+i,argv[0],argv[1],1);
                argv+=2;
        }
        cf[n].file=NULL; // end marker, makes some algorithms nicer

        // sort files by size for quicker "hole searching" 
        qsort(cf,n,sizeof(*cf),(int(*)(const void*,const void*))compare_size);

        // check if there are enough free partitions.
        if(!find_partitions(sb,cf,force)) {
                if(force) die(OUT_OF_PARTITIONS,"not enough free partitions");
                else die(NEED_FORCE,"not enough partitions, try using --force");
        }


        h=find_holes(sb);
        if(append) {
                for(i=0;h[i].start;i++);
                if(h[i-1].end==sb->size-1) {
                        h[i-1].end+=space_needed(cf);
                        h[i-1].unused+=space_needed(cf);
                } else {
                        h[i].start=sb->size;
                        h[i].end=sb->size+space_needed(cf); 
                        h[i].unused=space_needed(cf);
                }
        }


#ifdef DEBUG
        for(i=0;h[i].start;i++)
                printf("Hole %d: 0x%08x - 0x%08x (%d bytes)\n",i,h[i].start,h[i].end,h[i].unused);
#endif
        if(free_space(h)<space_needed(cf)) {
                if(force) die(PARTITION_TOO_SMALL,"not enough space");
                else die(NEED_FORCE,"not enough space, try using --force");
        }

        // if there is only one hole, and two images, and the
        // active image starts at the beginning, create an artificial
        // hole at the end the the packcramfs partition
        if(!h[1].start && sb->images==2 && !(sb->image[0].active*sb->image[1].active) && sb->image[0].active+sb->image[1].active
                && h[0].start!=START(sb) && n==1 && h[0].end-h[0].start+1>=cf[0].size) 
                        h[0].start=h[0].end-cf[0].size+1;


        if(!drop_cramfs_into_holes(cf,h)) {
                if(force) {
                        compress_all(f,sb);
                        superblock_written=1;
                        // drop... should always work at this point
                        h=find_holes(sb);
                        if(!drop_cramfs_into_holes(cf,h)) 
                                die(PARTITION_TOO_SMALL,"internal error 14");
                } else die(NEED_FORCE,"cannot fit cramfs images in, try using --force");
        }
#ifdef DEBUG
        for(i=0;cf[i].file!=NULL;i++)
                printf("Part. %d: %s --> hole %d\n",cf[i].partition,cf[i].mountpoint,cf[i].hole);
#endif

        // write superblock out if force is used
        if(force && !superblock_written) write_superblock(f,sb);


        for(i=0;cf[i].file;i++) {
                if(!quiet) printf("writing cramfs %s\n",cf[i].mountpoint);
                write_cramfs(f,sb,cf+i,h+cf[i].hole);
        }

        fclose(f);
        // check CRC of new images (close and reopen file to delete at least
        // part of the buffers involved 
        f=opendev(dev,MC_UPDATE,NULL);
        for(i=0;cf[i].file;i++)
                if(!check_cramfs_crc(f,cf[i].start,cf[i].size))
                        die(INTERNAL_ERROR,"write error: corrupt flash");

        if(append) {
                u32 last=0;
                for(i=0;cf[i].file;i++)
                        if(last<cf[i].start+cf[i].size)
                                last=cf[i].start+cf[i].size;
                if(last>sb->size) sb->size=last;
        }

        write_superblock(f,sb);
        if(!quiet) print_superblock(sb,1,terse);
        return 0;
}

/*
 * Extract specific image.
 */
static int extract_image(FILE *f,struct super_block *sb,int image,char *file) {
        FILE *n;
        char buf[1024];
        int i,l;
        if(file==NULL || !strcmp(file,"-"))
                n=stdout;
        else 
                n=fopen(file,"w");
        if(n==NULL) die(ILLEGAL_ARGUMENT,"cannot open output file");

        if(image<0 || image>sb->images || sb->image[image].start<=0 || sb->image[image].len<=0)
                die(ILLEGAL_ARGUMENT,"partition error");
        fseek(f,sb->image[image].start,SEEK_SET);
        l=sb->image[image].len;

        while(l && (i=fread(buf,1,l>1024?1024:l,f))>0) {
                l-=i;
                if(i!=fwrite(buf,1,i,n)) 
                        die(WRITE_ERROR,"cannot write to file");
        }
        if(l) die(CORRUPT_FILES,"error while reading packed image");

        return 0;
        
}

/*
 * Extract an image to a file or stdout (if there is no file 
 * or the filename is '-').
 */
static int extract(u8 *dev,char *mt,char *file) {
        int i,j;
        FILE *f;
        struct super_block *sb;
        f=opendev(dev,MC_READ,&sb);
        j=-1;
        for(i=0;i<sb->images;i++) {
                if(!strcmp(sb->image[i].mountpoint,mt)) {
                        if(sb->image[i].active) return extract_image(f,sb,i,file);
                        if(j<0) j=i;
                }
        }
        if(j>=0) return extract_image(f,sb,j,file);
        if(mt[0]>='0' && mt[0]<='9') {
                i=atoi(mt);
                if(i>=0 && i<sb->images) return extract_image(f,sb,i,file);
        }
        die(ILLEGAL_ARGUMENT,"arguments are wrong");
        return -1;
}

/*
 * Reactive all partitions that are not active and
 * delete all active partitions that have the same
 * mountpoint as a reactivated one.
 */
static void restore(char * dev) {
        FILE *f;
        int i,j;
        struct super_block *sb;
        f=opendev(dev,MC_UPDATE,&sb);
        do_umount_all(f,sb);
        for(i=0;i<sb->images;i++)
                if(!sb->image[i].active && PARTITION_USED(sb->image[i])) {
                        if(!quiet) printf("reactivating partition %d\n",i);
                        for(j=0;j<sb->images;j++)
                                if(sb->image[j].active==1 && !strcmp(sb->image[i].mountpoint,sb->image[j].mountpoint))
                                        sb->image[j].active=3;
                        sb->image[i].active=2;
                }
        for(i=0;i<sb->images;i++)
                if(sb->image[i].active==2) 
                        sb->image[i].active=1;
                else if(sb->image[i].active==3) 
                        sb->image[i].active=0;
        write_superblock(f,sb);
        print_superblock(sb,1,0);
}

/*
 * Deactive a partition (and delete all other partition
 * with the same mountpoint that are not active).
 */
static void deactivate(char * dev,char * part) {
        int digit=1,i;
        FILE *f;
        struct super_block *sb;
        f=opendev(dev,MC_UPDATE,&sb);
        do_umount_all(f,sb);

        digit=atoi(part);
        if(digit!=0 || part[0]=='0') {
                if(digit>=sb->images) die(ILLEGAL_ARGUMENT,"partition does not exist");
                if(!sb->image[digit].active) die(ILLEGAL_ARGUMENT,"partition already deactivated");
                for(i=0;i<sb->images;i++) 
                        if(i!=digit && !strcmp(sb->image[i].mountpoint,sb->image[digit].mountpoint))
                                delete_partition(&sb->image[i]);
                sb->image[digit].active=0;
        } else {
                int found=0;
                for(i=0;!found && i<sb->images;i++) 
                        if(!strcmp(sb->image[i].mountpoint,part)) 
                                found=sb->image[i].active;
                if(!found) die(ILLEGAL_ARGUMENT,"no active partition found");
                for(i=0;i<sb->images;i++) 
                        if(!strcmp(sb->image[i].mountpoint,part)) {
                                if(sb->image[i].active)
                                        sb->image[i].active=0;
                                else
                                        delete_partition(&sb->image[i]);
                        }
        }
        write_superblock(f,sb);
        print_superblock(sb,1,0);
}

/*
 * Activate the partitions 
 */
static void activate(char *dev,char *part[]) {
        int i;
        FILE *f;
        char **p;
        struct super_block *sb;
        f=opendev(dev,MC_UPDATE,&sb);
        do_umount_all(f,sb);

        p=part;
        if(p==NULL || *p==NULL) return;

        do {
                int found_active=0;
                int found_passive=0;
                for(i=0;i<sb->images;i++) {
                        if(!strcmp(*p,sb->image[i].mountpoint)) {
                                if(sb->image[i].active) {
                                        if(found_active) {
                                                fprintf(stderr,"found two active partitions with name %s\n",*p);
                                                die(CORRUPT_SUPERBLOCK,"filesystem error");
                                        }
                                        found_active=1;
                                } else {
                                        if(found_passive) {
                                                fprintf(stderr,"found two passive partitions with name %s\n",*p);
                                                die(CORRUPT_SUPERBLOCK,"filesystem error");
                                        }
                                        found_passive=1;
                                }
                        }
                }
                if(!found_passive) {
                        fprintf(stderr,"activation failed: partition %s not found\n",*p);
                        die(ILLEGAL_ARGUMENT,"cannot find one partition, nothing has been activated, aborting");
                }
        } while(*++p);

        p=part;
        do {
                for(i=0;i<sb->images;i++) 
                        if(!strcmp(*p,sb->image[i].mountpoint)) 
                                sb->image[i].active=1-sb->image[i].active;
        } while(*++p);

        write_superblock(f,sb);
        if(!quiet) print_superblock(sb,1,0);
}


#ifndef NO_DIRECT_MOUNT
/*
 * Mount via creation of a partition.
 */
static void mount_direct(FILE *f,struct super_block *sb,char *dev,char *root,int part,FILE *disk,unsigned long offset) {
        int minor,major;
        struct stat statbuf;
        char diskpath[128];
        char dir[PATH_MAX];

        if(fstat(fileno(f), &statbuf)) 
                die(ILLEGAL_ARGUMENT,"cannot stat() device");
        if(!S_ISBLK(statbuf.st_mode)) 
                die(ILLEGAL_ARGUMENT,"need a block device!");
        major=statbuf.st_rdev>>8;
        minor=part+sb->loopoffset+5;

        //printf("Creating partition for image %d - %s\n",part,sb->image[part].mountpoint);
        sprintf(diskpath,DISKPATH "_%x_%x",major,minor);
        if(stat(diskpath, &statbuf)) {
                mknod(diskpath,S_IFBLK|0600,(major<<8)+minor);
                if(!stat(diskpath, &statbuf) == 0) 
                        die(ILLEGAL_ARGUMENT,"cannot create partition device");
        }

        if(add_part(fileno(disk),minor,offset+sb->image[part].start/512,sb->image[part].len/512))
                die(INTERNAL_ERROR,"Cannot create kernel partition");
        if(strlen(root)+strlen(sb->image[part].mountpoint)+1>PATH_MAX) 
                die(ILLEGAL_ARGUMENT,"mount path too long");
        strcpy(dir,root);
        if(strcmp(sb->image[part].mountpoint,"/")) {
                if(sb->image[part].mountpoint[0]!='/') strcat(dir,"/");
                strcat(dir,sb->image[part].mountpoint);
        }

        //printf("mount(%s,%s)\n",diskpath,dir);
        if(mount(diskpath, dir, "cramfs", MS_RDONLY|MS_MGC_VAL, NULL)) {
                perror("mount");
                die(BUSY,"error while mounting cramfs image");
        }
}
#endif

#ifndef NO_LOOPBACK
/*
 * set loopback device (stolen from busybox)
 */
static int mount_part(FILE *f,struct super_block *sb,char *dev,char *root,int part) {
        struct loop_info loopinfo;
        int fd;
        char loopdev[strlen(LOOPPATH)+5];
        struct stat statbuf;
        char dir[PATH_MAX];

        sprintf(loopdev,LOOPPATH "%d",part+sb->loopoffset);

        strcpy(dir,root);
        if(strcmp(sb->image[part].mountpoint,"/")) {
                if(sb->image[part].mountpoint[0]!='/') strcat(dir,"/");
                strcat(dir,sb->image[part].mountpoint);
        }

        if(!quiet) printf("mounting %s, offset %d, as %s, on %s\n",sb->image[part].mountpoint,sb->image[part].start,loopdev,dir);

        if(stat(loopdev, &statbuf)) {
                mknod(loopdev,S_IFBLK|0600,0x700+part+sb->loopoffset);
                if(!stat(loopdev, &statbuf) == 0) 
                        die(ILLEGAL_ARGUMENT,"cannot create loop device");
        }

        if( S_ISBLK(statbuf.st_mode) && (fd = open(loopdev, O_RDONLY)) >= 0) {
                if (ioctl(fd, LOOP_GET_STATUS, &loopinfo) == -1) {
                        if (errno != ENXIO) {   /* probably free */
                                die(BUSY,"loopback device cannot be used");
                        }
                } else {
                        ioctl(fd, LOOP_CLR_FD, 0);
                }
                close(fd);
        }

        if((fd=open(loopdev,O_RDWR))<0) {
                perror(loopdev);
                die(BUSY,"error opening loopback device");
        }

        memset(&loopinfo, 0, sizeof(loopinfo));
        strncpy(loopinfo.lo_name, dev, LO_NAME_SIZE);
        loopinfo.lo_name[LO_NAME_SIZE - 1] = 0;
        (u32)loopinfo.lo_rdevice=sb->image[part].start;
        //(u32)loopinfo.lo_offset=sb->image[part].start;
        loopinfo.lo_encrypt_key_size = 0;
        if(ioctl(fd, LOOP_SET_FD, fileno(f))<0) {
                perror("ioctl: LOOP_SET_FD");
                die(BUSY,"error setting up loop device");
        }
        if (ioctl(fd, LOOP_SET_STATUS, &loopinfo) < 0) {
                perror("ioctl: LOOP_SET_STATUS");
                (void) ioctl(fd, LOOP_CLR_FD, 0);
                die(BUSY,"error setting up loop device");
        }

        if(strlen(root)+strlen(sb->image[part].mountpoint)+1>PATH_MAX) 
                die(ILLEGAL_ARGUMENT,"mount path too long");

        if(mount(loopdev, dir, "cramfs", MS_RDONLY|MS_MGC_VAL, NULL)) {
                perror("mount");
                die(BUSY,"error while mounting cramfs image");
        }
        return 0;
}
#endif

/*
 * sort mount data array by length of mointpoint.
 */
static int compare_len(struct umount_data *a,struct umount_data *b) {
        return -compare_rlen(a,b);
}

/*
 * Mount all partitions.
 */
static void mount_all(u8 *dev,u8 *root) {
        struct umount_data *ud;
        int i;
        struct super_block *sb;
        FILE *f,*disk;
        int use_direct=1;
        unsigned long offset;
        f=opendev(dev,MC_READ,&sb);

        // just to be on the save side
        do_umount_all(f,sb);

        // need to mount in correct order to avoid mount errors because of non
        // existing mountpoints.
        ud=(struct umount_data *)malloc(sizeof(struct umount_data)*sb->images);
        for(i=0;i<sb->images;i++) {
                ud[i].part=i;
                ud[i].name=sb->image[i].mountpoint;
        }

        qsort(ud,sb->images,sizeof(*ud),(int(*)(const void*,const void*))compare_len);

        for(i=0;i<sb->images;i++) 
                if(ud[0].name[0]!=0) {
                        if(strcmp(ud[0].name,"/")) {
                                fprintf(stderr,"first partition: %s\n",ud[0].name);
                                die(CORRUPT_FILES,"there is no root partition!");
                        }
                        break;
                }

#ifndef NO_DIRECT_MOUNT
        use_direct=check_for_direct_mount(f,sb,&disk,&offset);
#endif


        for(i=0;i<sb->images;i++)
                if(sb->image[ud[i].part].active) {
#ifndef NO_DIRECT_MOUNT
                        if(use_direct)
                                mount_direct(f,sb,dev,root,ud[i].part,disk,offset);
                        else
#endif
#ifdef NO_LOOPBACK
                                die(INTERNAL_ERROR,"Loopback support needed");
#else
                                mount_part(f,sb,dev,root,ud[i].part);
#endif
                }
}

/*
 * Unount all partitions.
 */
static void umount_all(u8 *dev) {
        struct super_block *sb;
        FILE *f;
        f=opendev(dev,MC_READ,&sb);

        // just to be on the save side
        do_umount_all(f,sb);
}

/*
 * Check one file
 */
static void check_image(u8 *dev) {
        struct cramfs_file cf;
        cf.size=0;
        
        read_cramfs(&cf,"",dev,1);
        printf("edition %u  files %u  blocks %u\n",cf.super.fsid.edition,cf.super.fsid.files,cf.super.fsid.blocks);
}

/*
 * Check all partitions.
 */
static void check(u8 *dev,int all,int write,int resize,int fast) {
        int i,error=0;
        struct super_block *sb;
        struct stat statbuf;
        FILE *f;
        if(write) f=opendev(dev,MC_UPDATE,&sb);
        else f=opendev(dev,MC_READ,&sb);

        // superblock has been checked already (and maybe fixed)
        for(i=0;i<sb->images && !fast;i++) {
                if((all || sb->image[i].active) && sb->image[i].start) {
                        if(!check_cramfs_crc(f,sb->image[i].start,sb->image[i].len)) {
                                error=CHECK_ERROR;
                                if(write) {
                                        fprintf(stderr,"partition %d %s deleted due to crc error\n",i,sb->image[i].mountpoint);
                                        delete_partition(&sb->image[i]);
                                } else fprintf(stderr,"partition %d %s should be deleted due to crc error\n",i,sb->image[i].mountpoint);
                        }
                }
        }
        if(!error && resize && write) {
                u32 s;
                fseek(f,0,SEEK_END);
                if(fstat(fileno(f), &statbuf)) die(INTERNAL_ERROR,"cannot stat() device");
                if(S_ISBLK(statbuf.st_mode)) {
			if(ioctl(fileno(f),BLKGETSIZE,&s) || s==0) 
				die(ILLEGAL_ARGUMENT,"cannot determine maximum file size");
			s*=512;
		} else { 
                        s=0;
                        for(i=0;i<sb->images;i++)
                                if(sb->image[i].start+sb->image[i].len>s)
                                        s=sb->image[i].start+sb->image[i].len;
                }
                if(s<=sb->size) {
                        for(i=0;i<sb->images;i++)
                                if(sb->image[i].start+sb->image[i].len>s)
                                        die(PARTITION_TOO_SMALL,"there is a partition that extends past the end of the requested size.");
                        ftruncate(fileno(f),s);
                } else {
                        fseek(f,s-1,SEEK_SET);
                        fwrite(&s,1,1,f);
                }
                sb->size=s;
                write_superblock(f,sb);
        }

        if(error) {
                if(write) {
                        write_superblock(f,sb);
                        die(error,"Some partitions were deleted");
                        print_superblock(sb,1,0);
                }
                die(error,"some partitions need to be deleted");
        }
        exit(error_code);
}


#define ARG_PRINT      'p'
#define ARG_CREATE     'c'
#define ARG_DEACTIVATE 'D'
#define ARG_STORE      's'
#define ARG_EXTRACT    'x'
#define ARG_ACTIVATE   'A'
#define ARG_MOUNT      'm'
#define ARG_UMOUNT     'u'
#define ARG_CHECK      'C'
#define ARG_COMPRESS   'z'
#define ARG_RESTORE    'r'
#define ARG_QUIET      'q'
#define ARG_WRITE      'w'
/*
 * main
 * See usage for description of options
 *
 */
int main(int argc,char *argv[]) {
        int work=0;
        int images=16;
        int all=0;
        int force=0;
        int write=0;
        int blocks=0;
        int blocksize=1024;
        int resize=0;
        int terse=0;
        int append=0;
        int image=0;
        int fast=0;
        int loopoffset=0;
        static struct option long_options[] = {
                { "create",0,0, ARG_CREATE },
                { "images",1,0,'i' },
                { "print",0,0, ARG_PRINT },
                { "all",0,0,'a' },
                { "terse",0,0,'t' },
                { "deactivate",0,0, ARG_DEACTIVATE },
                { "store",0,0, ARG_STORE },
                { "extract",0,0, ARG_EXTRACT },
                { "fast",0,0,'3' },
                { "force",0,0,'f' },
                { "image",0,0,'I' },
                { "activate",0,0, ARG_ACTIVATE },
                { "mount",0,0, ARG_MOUNT },
                { "umount",0,0, ARG_UMOUNT },
                { "loop",1,0, 'l' },
                { "check",0,0, ARG_CHECK },
                { "blocksize",1,0,'b' },
                { "compress",0,0, ARG_COMPRESS },
                { "restore",0,0, ARG_RESTORE },
                { "quiet",0,0, ARG_QUIET },
                { "write",0,0, ARG_WRITE },
                { "resize",0,0, '1' },
                { "append",0,0, '2' },
                { "progress-bar",0,0, 'P' },
                { 0,0,0,0 }
        };
        if(argc==1) usage(NULL);
        while(1) {
                int c=getopt_long(argc,argv,"ci:b:l:prq0aDtPIxwsfAzumC",long_options,NULL);
                if(c<0) break;
                switch(c) {
                case ARG_PRINT:
                case ARG_CREATE:
                case ARG_DEACTIVATE:
                case ARG_STORE:
                case ARG_ACTIVATE:
                case ARG_MOUNT:
                case ARG_UMOUNT:
                case ARG_COMPRESS:
                case ARG_RESTORE:
                case ARG_CHECK:
                case ARG_EXTRACT:
                        if(work) usage("only one of -p -c -D -A -s -x -m -u -z -r -C can be used");
                        work=c;
                        break;
                case 'i': images=atoi(optarg);
                          if(images<0 || images>128) usage("number of partitions requested is out of range (1<128)");
                          break;
                case 'P': progress_bar=1;break;
                case 'a': all=1; break;
                case 'w': write=1; break;
                case 'f': force=1; break;
                case 'l': loopoffset=atoi(optarg); 
                          if(loopoffset<0 || loopoffset>255) usage("loopoffset out of range");
                          break;
                case 'b': blocksize=atoi(optarg); 
                          if(blocksize<16 || blocksize>0x1000) usage("blocksize out of range");
                          break;
                case 'q': quiet=1;break;
                case 't': terse=1;break;
                case 'I': image=1;break;
                case '1': resize=1;break;
                case '2': append=1;break;
                case '3': fast=1;break;

                default: 
                          fprintf(stderr,"argument %c unknown\n",c);
                          usage("unknwon argument used\n");break;
                }
        }
        if(!work) usage("nothing to do?");
        switch(work) {
        case ARG_CHECK:
                if(optind!=argc-1) usage("wrong number of arguments");
                if(!write && resize) usage("--resize works only if used together with -w");
                if(image)
                        check_image(argv[optind]);
                else
                        check(argv[optind],all,write,resize,fast);
                break;
        case ARG_PRINT: 
                if(optind!=argc-1) usage("wrong number of arguments");
                if(image) 
                        print_image(argv[optind]);
                else
                        print(argv[optind],all,terse);
                break;
        case ARG_RESTORE: 
                if(optind!=argc-1) usage("wrong number of arguments");
                restore(argv[optind]);
                break;
        case ARG_CREATE:
                if(optind!=argc-2) usage("wrong number of arguments");
                blocks=atoi(argv[optind+1]);
                if(blocks<1 || blocks>0x100000) usage("number of blocks out of range");
                create(argv[optind],blocks,blocksize,images,loopoffset);
                break;
        case ARG_STORE:
                if(optind>argc-2) usage("wrong number of arguments");
                store(argv[optind],argv+optind+1,force,append,terse);
                break;
        case ARG_EXTRACT:
                if(optind!=argc-2 && optind!=argc-3) usage("wrong number of arguments");
                extract(argv[optind],argv[optind+1],optind==argc-3?argv[optind+2]:NULL);
                break;
        case ARG_COMPRESS:
                if(optind!=argc-1) usage("wrong number of arguments");
                compress_mc(argv[optind]);
                break;
        case ARG_DEACTIVATE:
                if(optind!=argc-2) usage("wrong number of arguments");
                deactivate(argv[optind],argv[optind+1]);
                break;
        case ARG_ACTIVATE:
                if(optind>argc-2) usage("wrong number of arguments");
                activate(argv[optind],argv+optind+1);
                break;
        case ARG_MOUNT:
                if(optind!=argc-2) usage("wrong number of arguments");
                mount_all(argv[optind],argv[optind+1]);
                break;
        case ARG_UMOUNT:
                if(optind!=argc-1) usage("wrong number of arguments");
                umount_all(argv[optind]);
                break;
        default: die(ILLEGAL_ARGUMENT,"this feature is not yet implemented");break;
        }
        exit(error_code);
}
